==> ja-WiKicker-0.23.tbz: lib/perl5/site_perl/5.8.5/WiKicker/DB/File.pm <==
#
# WiKicker
#
#  File.pm
#
#
# Copyright (C) 2003,2004 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: File.pm,v 1.13 2004/04/02 12:17:03 naney Exp $
#
#

package WiKicker::DB::File;

use 5.005;
use strict;
use WiKicker::Loader;
use Storable qw(freeze thaw);
use WiKicker::Tie::KeyLockFile;
use WiKicker::Tie::ExternalAllLockFile;
use WiKicker::Tie::ExternalAllLockRcs;
use WiKicker::Lock;
use Fcntl ':flock';
use AutoLoader 'AUTOLOAD';
use Carp;

use vars qw(@ISA);
@ISA = qw(WiKicker::Lock);

# my %instances;
# sub get_instance {
#   my $directory = shift;

#   croak "No directory name" unless $directory;

#   unless (exists $instances{$directory}) {
#     my $db = WiKicker::DB::File->new($directory);
#     $instances{$directory} = $db;
#   }
#   return $instances{$directory};
# }

sub new {
  my ($class, $directory) = @_;

  croak "No directory name" unless $directory;

  my $self = bless { directory => $directory }, $class;

  $self->_initialize;

  return $self;
}

sub _initialize {
  my $self = shift;

  unless (-d $self->{directory}) {
    WiKicker::Loader::load_module('File::Path');
    File::Path::mkpath([ $self->{directory} ], 0, 0777);
  }
}

sub DESTROY {
  my $self = shift;
  #-IFDEBUG: WiKicker::total_size("DB(before DESTROY)", $self);
  #-IFDEBUG: WiKicker::cycle_check($self);

  # @todo undef db
  # @todo untie db_hash
  $self->unlock_pages;
}

sub exists_live_page {
  my ($self, $key) = @_;

  return 0 unless $key;
  return exists $self->_page_db->{$key};
}

sub exists_backup_page {
  my ($self, $key) = @_;

  return 0 unless $key;
  return $self->_page_db_object->exists_backup($key);
}

sub page {
  my ($self, $key, $value, $author, $subject) = @_;

  croak "No key" unless $key;

  if (defined $value) {
    if ($self->lock_mode_key('page') != LOCK_EX) {
      croak "no exclusive lock to write page";
    }
    $self->_page_db->{$key} = $value;

    $subject = $subject || '';
    my $revision_no = $self->_page_db_object->latest_revision_no($key);

    $self->information($key, 'lastmodified', time);
    $self->information($key, 'revisionno', $revision_no);
    $self->information($key, 'author', $author);
    $self->information($key, 'subject', $subject);
    $self->information($key, 'lastcomment', '');
    $self->_update_recent_changes($key, time, $author, $subject);
    $self->_touch_last_modified_file;
    return $value;
  }
  else {
    return $self->_page_db->{$key};
  }
}


sub page_revision_no {
  my ($self, $key) = @_;

  croak "No key" unless $key;


  unless ($self->lock_mode_key('page')) {
    croak "no lock to get revision no";
  }

  unless ($self->exists_live_page($key)) {
    return 0;
  }
  return $self->information($key, 'revisionno');
}

sub get_revision {
  my ($self, $key, $version) = @_;

  croak "No key" unless $key;

  return $self->_page_db_object->get_revision($key, $version);
}

sub get_head_revision {
  my ($self, $key) = @_;

  croak "No key" unless $key;

  return $self->_page_db_object->get_head_revision($key);
}

sub page_rlog {
  my ($self, $key) = @_;

  croak "No key" unless $key;

  return $self->_page_db_object->log($key);
}

sub page_list {
  my $self = shift;

  #unless ($self->lock_mode_key('page')) {
  #  croak "no lock to get page list";
  #}

  #return keys %{$self->_page_db};
  return $self->_page_db_object->all_keys;
}

sub page_name_list {
  my $self = shift;

  if ($self->{page_name_list}) {
    return $self->{page_name_list};
  }
  return $self->{page_name_list} = WiKicker::PageNameList
    ->new($self->_page_db_object->all_keys_ref);
}

sub page_list_with_predicate {
  my ($self, $predicate) = @_;

  unless ($self->lock_mode_key('page')) {
    croak "no lock to get page list";
  }

  return $self->_page_db_object->all_keys_with_predicate($predicate);
}

sub ancestors_page_name_list {
  my ($self, $name) = @_;

  my @path = split('/', $name);
  my @result = ();
  my @path_name;
  for (my $i = 0; $i < @path - 1; $i++) {
    push(@path_name, $path[$i]);
    if ($self->exists_live_page(join('/', @path_name))) {
      push(@result, join('/', @path_name));
    }
  }

  return @result;
}

# basic
# lastmodified . epoch
#
sub information {
  my ($self, $key, $field, $value) = @_;

  if (defined $value) {
    if ($self->lock_mode_key('page') != LOCK_EX) {
      croak "No exclusive lock to write page information";
    }
  }

  unless (defined $value) {
    if (exists $self->{info_cache}->{$key}) {
      return $self->{info_cache}->{$key}->{$field};
    }
  }

  my $db = $self->_info_db;
  my %record;
  if (exists $db->{$key}) {
    my $data = $db->{$key};
    unless ($data) {
      return undef;
    }
    %record = %{thaw($data)};
    $self->{info_cache}->{$key} = \%record;
  }
  if (defined $value) {
    $record{$field} = $value;
    $self->{info_cache}->{$key}->{$field} = $value;
    $db->{$key} = freeze(\%record);
  }
  else {
    return $record{$field};
  }
}

sub last_modified {
  my ($self, $name) = @_;

  my $value = $self->information($name, 'lastmodified');
  if (!(defined $value) || $value eq '') {
    return '0';
  }
  else {
    return $value;
  }
}

sub page_subject {
  my ($self, $key) = @_;

  return $self->information($key, 'subject');
}

sub page_author {
  my ($self, $name) = @_;

  return $self->information($name, 'author');
}

sub cache {
  my ($self, $name, $key, $value) = @_;

  croak "No name" unless $name;
  croak "No key" unless $key;

  if (defined $value) {
    return $self->_cache_db($name)->{$key} = $value;
  }
  else {
    return $self->_cache_db($name)->{$key};
  }
}

sub delete_cache {
  my ($self, $name, $key) = @_;

  delete $self->_cache_db($name)->{$key};
}

sub exists_cache {
  my ($self, $name, $key) = @_;

  return exists $self->_cache_db($name)->{$key};
}

sub clear_cache {
  my ($self, $name) = @_;

  my $db = $self->_cache_db_object($name);

  $db->{CLOBER} = 2;
  $db->CLEAR;
}

sub list_cache {
  my ($self, $name) = @_;

  return $self->_cache_db_object($name)->all_keys;
}

sub page_log {
  my ($self, $name, $key, $value) = @_;

  my $db = $self->_page_log_db($name);
  if (defined $value) {
    return $self->_page_log_db($name)->{$key} = $value;
  }
  else {
    return $self->_page_log_db($name)->{$key};
  }
}

sub append_page_log {
  my ($self, $name, $key, $value) = @_;

  $self->_page_log_db_object($name)->append($key, $value);
}

sub _page_db {
  my $self = shift;

  if (exists $self->{page_db}) {
    return $self->{page_db};
  }

  my %db_hash;
  my $db = tie(%db_hash, 'WiKicker::Tie::ExternalAllLockRcs', $self->{directory} . '/page');
  $self->{page_db_object} = $db;
  return $self->{page_db} = \%db_hash;
}

sub _page_db_object {
  my $self = shift;

  if (exists $self->{page_db_object}) {
    return $self->{page_db_object};
  }

  my %db_hash;
  my $db = tie(%db_hash, 'WiKicker::Tie::ExternalAllLockRcs', $self->{directory} . '/page');
  $self->{page_db} = \%db_hash;
  return $self->{page_db_object} = $db;
}

sub _cache_db {
  my ($self, $name) = @_;

  return $self->_named_key_db($name, 'cache_db', 'cache');
}

sub _cache_db_object {
  my ($self, $name) = @_;

  return $self->_named_key_db_object($name, 'cache_db', 'cache');
}

sub _page_log_db {
  my ($self, $name) = @_;

  return $self->_named_key_db($name, 'page_log_db', 'pagelog');
}

sub _page_log_db_object {
  my ($self, $name) = @_;

  return $self->_named_key_db_object($name, 'page_log_db', 'pagelog');
}

sub _info_db {
  my ($self, $name) = @_;

  return $self->_named_db('basic', 'info_db', 'info');
}

sub _info_db_object {
  my ($self, $name) = @_;

  return $self->_named_db_object('basic', 'info_db', 'info');
}


sub _named_db {
  my ($self, $name, $hash_field, $directory) = @_;

  croak "No name" unless $name;

  if (exists $self->{hash}->{$hash_field}->{$name}) {
    return $self->{hash}->{$hash_field}->{$name};
  }

  my %db_hash;
  my $db = tie(%db_hash, 'WiKicker::Tie::ExternalAllLockFile', $self->{directory} . "/$directory/$name");
  $self->{db_obj}->{$hash_field}->{$name} = $db;
  return $self->{hash}->{$hash_field}->{$name} = \%db_hash;
}

sub _named_db_object {
  my ($self, $name, $hash_field, $directory) = @_;

  $self->_named_db($name, $hash_field, $directory); # open
  return $self->{db_obj}->{$hash_field}->{$name};
}

sub _named_key_db {
  my ($self, $name, $hash_field, $directory) = @_;

  croak "No name" unless $name;

  if (exists $self->{hash}->{$hash_field}->{$name}) {
    return $self->{hash}->{$hash_field}->{$name};
  }

  my %db_hash;
  my $db = tie(%db_hash, 'WiKicker::Tie::KeyLockFile', $self->{directory} . "/$directory/$name");
  $self->{db_obj}->{$hash_field}->{$name} = $db;
  return $self->{hash}->{$hash_field}->{$name} = \%db_hash;
}

sub _named_key_db_object {
  my ($self, $name, $hash_field, $directory) = @_;

  $self->_named_key_db($name, $hash_field, $directory); # open
  return $self->{db_obj}->{$hash_field}->{$name};
}


sub lock_file_name {
  my ($self, $lock) = @_;

  return $self->{directory} . "/$lock.lock";
}

sub last_modified_file_name {
  my $self = shift;

  return $self->{directory} . '/last-modified.txt';
}

sub db_last_modified_time {
  my $self = shift;

  my @stat = stat($self->last_modified_file_name);
  unless (@stat) {
    return 0;
  }
  else {
    return $stat[9];
  }
}

sub shared_lock_pages {
  my $self = shift;

  $self->shared_lock_key('page');
  $self->_page_db_object->shared_lock_all_by_external;
  $self->_info_db_object->shared_lock_all_by_external;
}

sub exclusive_lock_pages {
  my $self = shift;

  $self->exclusive_lock_key('page');
  $self->_page_db_object->exclusive_lock_all_by_external;
  $self->_info_db_object->exclusive_lock_all_by_external;
}

sub unlock_pages {
  my $self = shift;

  $self->_info_db_object->unlock_all_by_external;
  $self->_page_db_object->unlock_all_by_external;
  $self->unlock_key('page');
}

sub delete_page;
sub _update_recent_changes;
sub _touch_last_modified_file;

1;

__END__
sub delete_page {
  my ($self, $key) = @_;

  croak "No key" unless $key;

  if ($self->lock_mode_key('page') != LOCK_EX) {
    croak "no exclusive lock to delete page";
  }

  delete $self->_page_db->{$key};
  $self->_touch_last_modified_file;
}

sub _update_recent_changes {
  my ($self, $key, $time, $author, $subject) = @_;

  my $recent_changes = $self->information('RecentChanges', 'changes');

  my @new_data;
  my $i = 0;
  for my $record (@{$recent_changes}) {
    $i++;
    if ($record->[0] ne $key) {
      push(@new_data, $record);
    }
    last if $i >= 50;
  }
  unshift(@new_data, [$key, $time, $author, $subject]);

  $self->information('RecentChanges', 'changes', \@new_data);
}

sub _touch_last_modified_file {
  my $self = shift;
  my $touch_file = $self->last_modified_file_name;
  unless(open(OUT, ">$touch_file")) {
    warn "Can't open touch file: $touch_file: $!";
    return;
  }
  print OUT time;
  close(OUT);
}

==> ja-WiKicker-0.26.tbz: lib/perl5/site_perl/5.8.9/WiKicker/DB/File.pm <==
#
# WiKicker
#
#  File.pm
#
#
# Copyright (C) 2003,2004,2005 WATANABE Yoshimasa <naney@naney.org>
#
# This file is part of the WiKicker.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
# $Id: File.pm,v 1.16 2005/05/06 13:06:09 naney Exp $
#
#

package WiKicker::DB::File;

use 5.005;
use strict;
use WiKicker::Loader;
use Storable qw(freeze thaw);
use WiKicker::Tie::KeyLockFile;
use WiKicker::Tie::ExternalAllLockFile;
use WiKicker::Tie::ExternalAllLockRcs;
use WiKicker::Lock;
use WiKicker::PageNameList;
use Fcntl ':flock';
use AutoLoader 'AUTOLOAD';
use Carp;
use File::Spec;

use vars qw(@ISA);
@ISA = qw(WiKicker::Lock);

sub new {
  my ($class, $directory) = @_;

  croak "No directory name" unless $directory;

  my $self = bless { directory => $directory }, $class;

  $self->_initialize;

  return $self;
}

sub DESTROY {
  my $self = shift;
  #-IFDEBUG: WiKicker::total_size("DB(before DESTROY)", $self);
  #-IFDEBUG: WiKicker::cycle_check($self);

  # @todo undef db
  # @todo untie db_hash
  $self->unlock_pages;
}

sub exists_live_page {
  my ($self, $key) = @_;

  return 0 unless defined $key;
  return exists $self->_page_db->{$key};
}

sub exists_backup_page {
  my ($self, $key) = @_;

  return 0 unless defined $key;
  return $self->_page_db_object->exists_backup($key);
}

# require
#  shared_lock_pages (read)
#  exclusive_lock_pages (write)
sub page {
  my ($self, $key, $value, $author, $subject) = @_;

  croak 'No key' unless $key;

  return $self->_write_page($key, $value, $author, $subject) if defined $value;
  return $self->_page_db->{$key};
}


# require
# shared_lock_pages
sub page_revision_no {
  my ($self, $key) = @_;

  croak 'No key' unless $key;
  croak 'no lock to get revision no' unless $self->lock_mode_key('page');
  return '0' unless ($self->exists_live_page($key));
  return $self->information($key, 'revisionno');
}

sub get_revision {
  my ($self, $key, $version) = @_;

  croak 'No key' unless $key;
  return $self->_page_db_object->get_revision($key, $version);
}

sub get_head_revision {
  my ($self, $key) = @_;

  croak 'No key' unless $key;
  return $self->_page_db_object->get_head_revision($key);
}

sub page_list {
  my $self = shift;

  return $self->_page_db_object->all_keys;
}

sub page_name_list {
  my $self = shift;

  return $self->{page_name_list} if $self->{page_name_list};

  return $self->{page_name_list} = WiKicker::PageNameList
    ->new($self->_page_db_object->all_keys_ref);
}

# basic
# lastmodified . epoch
#
sub information {
  my ($self, $key, $field, $value) = @_;

  if (defined $value) {
    $self->set_information($key, $field, $value);
    return;
  }

  if (exists $self->{info_cache}->{$key}) {
    return $self->{info_cache}->{$key}->{$field};
  }

  my $info_db = $self->_info_db;
  my %record;
  # get_record from db
  if (exists $info_db->{$key}) {
    my $data = $info_db->{$key};
    unless ($data) {
      return undef;
    }
    %record = %{thaw($data)};
    $self->{info_cache}->{$key} = \%record;
  }
  return $record{$field};
}

sub set_information {
  my ($self, $key, %data) = @_;

  $self->_check_page_ex_lock('No exclusive lock to write page information');

  my $info_db = $self->_info_db;
  my %record;
  # get record from db
  if (exists $info_db->{$key}) {
    my $serialized_object = $info_db->{$key};
    if ($serialized_object) {
      %record = %{thaw($serialized_object)};
    }
  }

  # update record (and cache)
  for my $field (keys %data) {
    $record{$field} = $data{$field};
  }

  $self->{info_cache}->{$key} = \%record;

  # write to db
  $info_db->{$key} = freeze(\%record);
}


sub last_modified {
  my ($self, $name) = @_;

  my $value = $self->information($name, 'lastmodified');
  if (!(defined $value) || $value eq '') {
    return '0';
  }
  else {
    return $value;
  }
}

# no record:  return undef
# no subject: return ''
sub page_subject {
  my ($self, $key) = @_;

  return $self->information($key, 'subject');
}

# no record:  return undef
# no subject: return ''
sub page_author {
  my ($self, $name) = @_;

  return $self->information($name, 'author');
}

sub cache {
  my ($self, $name, $key, $value) = @_;

  croak 'No name' unless $name;
  croak 'No key' unless $key;

  if (defined $value) {
    return $self->_cache_db($name)->{$key} = $value;
  }
  else {
    return $self->_cache_db($name)->{$key};
  }
}

sub exists_cache {
  my ($self, $name, $key) = @_;

  return exists $self->_cache_db($name)->{$key};
}


sub list_cache {
  my ($self, $name) = @_;

  return $self->_cache_db_object($name)->all_keys;
}

sub lock_file_name {
  my ($self, $lock) = @_;

  return File::Spec->catfile($self->{directory}, "$lock.lock");
}

sub last_modified_file_name {
  my $self = shift;

  return File::Spec->catfile($self->{directory}, 'last-modified.txt');
}

sub db_last_modified_time {
  my $self = shift;

  my @stat = stat($self->last_modified_file_name);
  unless (@stat) {
    return 0;
  }
  else {
    return $stat[9];
  }
}

sub shared_lock_pages {
  my $self = shift;

  $self->shared_lock_key('page');
  $self->_page_db_object->shared_lock_all_by_external;
  $self->_info_db_object->shared_lock_all_by_external;
}

sub exclusive_lock_pages {
  my $self = shift;

  $self->exclusive_lock_key('page');
  $self->_page_db_object->exclusive_lock_all_by_external;
  $self->_info_db_object->exclusive_lock_all_by_external;
}

sub unlock_pages {
  my $self = shift;

  $self->_info_db_object->unlock_all_by_external;
  $self->_page_db_object->unlock_all_by_external;
  $self->unlock_key('page');
}

sub _initialize {
  my $self = shift;

  unless (-d $self->{directory}) {
    WiKicker::Loader::load_module('File::Path');
    File::Path::mkpath([ $self->{directory} ], 0, 0777);
  }
}

sub _write_page {
  my ($self, $key, $value, $author, $subject) = @_;

  $self->_check_page_ex_lock('no exclusive lock to write page');

  # check page_name_list cache
  if ($self->{page_name_list} && ! $self->{page_name_list}->has($key)) {
    delete $self->{page_name_list};
  }

  $self->_page_db->{$key} = $value;

  $subject = $subject || '';
  $author = $author || '';
  my $revision_no = $self->_page_db_object->latest_revision_no($key);

  #$self->set_information($key, 'lastmodified', time);
  #$self->set_information($key, 'revisionno', $revision_no);
  #$self->set_information($key, 'author', $author);
  #$self->set_information($key, 'subject', $subject);
  #$self->set_information($key, 'lastcomment', '');
  $self->set_information($key,
                         'lastmodified', time,
                         'revisionno', $revision_no,
                         'author', $author,
                         'subject', $subject,
                         'lastcomment', '');

  $self->_update_recent_changes($key, time, $author, $subject);
  $self->_touch_last_modified_file;
  return $value;
}

sub _page_db {
  my $self = shift;

  return $self->{page_db} if exists $self->{page_db};
  $self->_init_page_db;
  return $self->{page_db};
}

sub _page_db_object {
  my $self = shift;

  return $self->{page_db_object} if exists $self->{page_db_object};
  $self->_init_page_db;
  return $self->{page_db_object};
}

sub _init_page_db {
  my $self = shift;

  my %db_hash;
  my $db;
  my $directory = File::Spec->catfile($self->{directory}, 'page');
  eval {
    $db = tie(%db_hash, 'WiKicker::Tie::ExternalAllLockRcs', $directory);
  };
  if ($@) {
    Log::Log4perl->get_logger->error($@);
    croak "Can't open Page DB '$directory' :$@";
  }
  $self->{page_db_object} = $db;
  $self->{page_db} = \%db_hash;
}

sub _cache_db {
  my ($self, $name) = @_;

  return $self->_named_key_db($name, 'cache_db', 'cache');
}

sub _cache_db_object {
  my ($self, $name) = @_;

  return $self->_named_key_db_object($name, 'cache_db', 'cache');
}

sub _info_db {
  my ($self, $name) = @_;

  return $self->_named_db('basic', 'info_db', 'info');
}

sub _info_db_object {
  my ($self, $name) = @_;

  return $self->_named_db_object('basic', 'info_db', 'info');
}


sub _named_db {
  my ($self, $name, $hash_field, $directory) = @_;

  croak 'No name' unless $name;

  if (exists $self->{hash}->{$hash_field}->{$name}) {
    return $self->{hash}->{$hash_field}->{$name};
  }

  my %db_hash;
  my $db;
  my $directory_path
    = File::Spec->catfile($self->{directory},
                          $directory,
                          $name);
  eval {
    $db = tie(%db_hash, 'WiKicker::Tie::ExternalAllLockFile', $directory_path);
  };
  if ($@) {
    Log::Log4perl->get_logger->error($@);
    croak "Can't open $name DB '$directory_path' :$@";
  }
  $self->{db_obj}->{$hash_field}->{$name} = $db;
  return $self->{hash}->{$hash_field}->{$name} = \%db_hash;
}

sub _named_db_object {
  my ($self, $name, $hash_field, $directory) = @_;

  $self->_named_db($name, $hash_field, $directory); # open
  return $self->{db_obj}->{$hash_field}->{$name};
}

sub _named_key_db {
  my ($self, $name, $hash_field, $directory) = @_;

  croak 'No name' unless $name;

  if (exists $self->{hash}->{$hash_field}->{$name}) {
    return $self->{hash}->{$hash_field}->{$name};
  }

  my %db_hash;
  my $db = tie(%db_hash, 'WiKicker::Tie::KeyLockFile',
               File::Spec->catfile($self->{directory},
                                   $directory,
                                   $name));
  $self->{db_obj}->{$hash_field}->{$name} = $db;
  return $self->{hash}->{$hash_field}->{$name} = \%db_hash;
}

sub _named_key_db_object {
  my ($self, $name, $hash_field, $directory) = @_;

  $self->_named_key_db($name, $hash_field, $directory); # open
  return $self->{db_obj}->{$hash_field}->{$name};
}



sub _check_page_ex_lock {
  my ($self, $message) = @_;

  croak $message if $self->lock_mode_key('page') != LOCK_EX;
}

sub delete_page;
sub _update_recent_changes;
sub _touch_last_modified_file;
sub delete_cache;
sub clear_cache;
sub page_rlog;

1;

__END__
sub delete_page {
  my ($self, $key) = @_;

  croak 'No key' unless $key;

  $self->_check_page_ex_lock('no exclusive lock to delete page');

  # check page_name_list cache
  if ($self->{page_name_list} && $self->{page_name_list}->has($key)) {
    delete $self->{page_name_list};
  }

  delete $self->_page_db->{$key};
  $self->_touch_last_modified_file;
}

sub _update_recent_changes {
  my ($self, $key, $time, $author, $subject) = @_;

  my $recent_changes = $self->information('RecentChanges', 'changes');

  my @new_data;
  my $i = 0;
  for my $record (@{$recent_changes}) {
    $i++;
    if ($record->[0] ne $key) {
      push(@new_data, $record);
    }
    last if $i >= 50;
  }
  unshift(@new_data, [$key, $time, $author, $subject]);

  $self->set_information('RecentChanges', 'changes', \@new_data);
}

sub _touch_last_modified_file {
  my $self = shift;
  my $touch_file = $self->last_modified_file_name;
  unless(open(OUT, ">$touch_file")) {
    warn "Can't open touch file: $touch_file: $!";
    return;
  }
  print OUT time;
  close(OUT);
}

sub delete_cache {
  my ($self, $name, $key) = @_;

  delete $self->_cache_db($name)->{$key};
}

sub clear_cache {
  my ($self, $name) = @_;

  my $db = $self->_cache_db_object($name);

  $db->{CLOBER} = 2;
  $db->CLEAR;
}

sub page_rlog {
  my ($self, $key) = @_;

  croak 'No key' unless $key;

  return $self->_page_db_object->log($key);
}
